import warnings


class Model:
    """A resource stores and manipulate objects in its attached storage.

    It is not aware of HTTP environment nor HTTP API.

    Objects are isolated according to the provided `name` and `parent_id`.

    Those notions have no particular semantic and can represent anything.
    For example, the resource `name` can be the *type* of objects stored, and
    `parent_id` can be the current *user id* or *a group* where the resource
    belongs. If left empty, the resource objects are not isolated.
    """

    id_field = "id"
    """Name of `id` field in objects"""

    modified_field = "last_modified"
    """Name of `last modified` field in objects"""

    deleted_field = "deleted"
    """Name of `deleted` field in deleted objects"""

    permissions_field = "__permissions__"
    # Permissions field used to annotate data with permissions.

    def __init__(
        self,
        storage,
        permission,
        id_generator=None,
        resource_name="",
        parent_id="",
        current_principal=None,
        prefixed_principals=None,
    ):
        """
        :param storage: an instance of storage
        :type storage: :class:`kinto.core.storage.Storage`
        :param id_generator: an instance of id generator, used by storage
            on object creation.

        :param str resource_name: the resource name
        :param str parent_id: the default parent id
        """
        self.storage = storage
        self.permission = permission
        self.id_generator = id_generator
        self.parent_id = parent_id
        self.resource_name = resource_name
        self.current_principal = current_principal
        self.prefixed_principals = prefixed_principals

        # Object permission id.
        self.get_permission_object_id = None

    def timestamp(self, parent_id=None):
        """Fetch the resource current timestamp.

        :param str parent_id: optional filter for parent id
        :rtype: int
        """
        parent_id = parent_id or self.parent_id
        return self.storage.resource_timestamp(
            resource_name=self.resource_name, parent_id=parent_id
        )

    def _annotate(self, obj, perm_object_id):
        permissions = self.permission.get_object_permissions(perm_object_id)
        # Permissions are not returned if user only has read permission.
        writers = permissions.get("write", [])
        principals = self.prefixed_principals + [self.current_principal]
        if len(set(writers) & set(principals)) == 0:
            permissions = {}
        # Insert the permissions values in the response.
        annotated = {**obj, self.permissions_field: permissions}
        return annotated

    def _allow_write(self, perm_object_id):
        """Helper to give the ``write`` permission to the current user.
        """
        self.permission.add_principal_to_ace(perm_object_id, "write", self.current_principal)

    def get_objects(
        self,
        filters=None,
        sorting=None,
        pagination_rules=None,
        limit=None,
        include_deleted=False,
        parent_id=None,
    ):
        """Fetch the resource objects.

        Override to post-process objects after feching them from storage.

        :param filters: Optionally filter the objects by their attribute.
            Each filter in this list is a tuple of a field, a value and a
            comparison (see `kinto.core.utils.COMPARISON`). All filters
            are combined using *AND*.
        :type filters: list of :class:`kinto.core.storage.Filter`

        :param sorting: Optionally sort the objects by attribute.
            Each sort instruction in this list refers to a field and a
            direction (negative means descending). All sort instructions are
            cumulative.
        :type sorting: list of :class:`kinto.core.storage.Sort`

        :param pagination_rules: Optionally paginate the list of objects.
            This list of rules aims to reduce the set of objects to the current
            page. A rule is a list of filters (see `filters` parameter),
            and all rules are combined using *OR*.
        :type pagination_rules: list of list of
            :class:`kinto.core.storage.Filter`

        :param int limit: Optionally limit the number of objects to be
            retrieved.

        :param bool include_deleted: Optionally include the deleted objects
            that match the filters.

        :param str parent_id: optional filter for parent id

        :returns: A list of objects in the current page.
        :rtype: list
        """
        parent_id = parent_id or self.parent_id
        return self.storage.list_all(
            resource_name=self.resource_name,
            parent_id=parent_id,
            filters=filters,
            sorting=sorting,
            pagination_rules=pagination_rules,
            limit=limit,
            include_deleted=include_deleted,
            id_field=self.id_field,
            modified_field=self.modified_field,
            deleted_field=self.deleted_field,
        )

    def count_objects(self, filters=None, parent_id=None):
        """Fetch the total count of resource objects

        Override to post-process objects after feching them from storage.

        :param filters: Optionally filter the objects by their attribute.
            Each filter in this list is a tuple of a field, a value and a
            comparison (see `kinto.core.utils.COMPARISON`). All filters
            are combined using *AND*.
        :type filters: list of :class:`kinto.core.storage.Filter`

        :param str parent_id: optional filter for parent id

        :returns: An integer of the total number of objects in the result set.
        :rtype: int
        """
        parent_id = parent_id or self.parent_id
        return self.storage.count_all(
            resource_name=self.resource_name,
            parent_id=parent_id,
            filters=filters,
            id_field=self.id_field,
            modified_field=self.modified_field,
            deleted_field=self.deleted_field,
        )

    def delete_objects(
        self, filters=None, sorting=None, pagination_rules=None, limit=None, parent_id=None
    ):
        """Delete multiple resource objects.

        Override to post-process objects after their deletion from storage.

        :param filters: Optionally filter the objects by their attribute.
            Each filter in this list is a tuple of a field, a value and a
            comparison (see `kinto.core.utils.COMPARISON`). All filters
            are combined using *AND*.
        :type filters: list of :class:`kinto.core.storage.Filter`

        :param sorting: Optionnally sort the objects by attribute.
            Each sort instruction in this list refers to a field and a
            direction (negative means descending). All sort instructions are
            cumulative.
        :type sorting: list of :class:`kinto.core.storage.Sort`

        :param pagination_rules: Optionnally paginate the deletion of objects.
            This list of rules aims to reduce the set of objects to the current
            page. A rule is a list of filters (see `filters` parameter),
            and all rules are combined using *OR*.
        :type pagination_rules: list of list of
            :class:`kinto.core.storage.Filter`

        :param int limit: Optionnally limit the number of objects to be
           deleted.

        :param str parent_id: optional filter for parent id

        :returns: The list of deleted objects from storage.
        """
        parent_id = parent_id or self.parent_id
        deleted = self.storage.delete_all(
            resource_name=self.resource_name,
            parent_id=parent_id,
            filters=filters,
            sorting=sorting,
            pagination_rules=pagination_rules,
            limit=limit,
            id_field=self.id_field,
            modified_field=self.modified_field,
            deleted_field=self.deleted_field,
        )
        # Take a huge shortcut in case we want to delete everything.
        if not filters:
            perm_ids = [self.get_permission_object_id(object_id="*")]
        else:
            perm_ids = [self.get_permission_object_id(object_id=r[self.id_field]) for r in deleted]
        self.permission.delete_object_permissions(*perm_ids)
        return deleted

    def get_object(self, object_id, parent_id=None):
        """Fetch current view related object, and raise 404 if missing.

        :param str object_id: object identifier
        :param str parent_id: optional filter for parent id

        :returns: the object from storage
        :rtype: dict
        """
        parent_id = parent_id or self.parent_id
        obj = self.storage.get(
            resource_name=self.resource_name,
            parent_id=parent_id,
            object_id=object_id,
            id_field=self.id_field,
            modified_field=self.modified_field,
        )
        perm_object_id = self.get_permission_object_id(object_id)
        return self._annotate(obj, perm_object_id)

    def create_object(self, obj, parent_id=None):
        """Create an object in the resource.

        The current principal is added to the owner (``write`` permission).

        Override to perform actions or post-process objects after their
        creation in storage.

        .. code-block:: python

            def create_object(self, obj):
                obj = super().create_object(obj)
                idx = index.store(obj)
                object['index'] = idx
                return object

        :param dict obj: object to store
        :param str parent_id: optional filter for parent id

        :returns: the newly created object.
        :rtype: dict
        """
        parent_id = parent_id or self.parent_id

        permissions = obj.pop(self.permissions_field, {})

        created = self.storage.create(
            resource_name=self.resource_name,
            parent_id=parent_id,
            obj=obj,
            id_generator=self.id_generator,
            id_field=self.id_field,
            modified_field=self.modified_field,
        )

        object_id = created[self.id_field]
        perm_object_id = self.get_permission_object_id(object_id)
        self.permission.replace_object_permissions(perm_object_id, permissions)
        self._allow_write(perm_object_id)

        return self._annotate(created, perm_object_id)

    def update_object(self, obj, parent_id=None):
        """Update object and the specified permissions.

        If no permissions is specified, the current permissions are not
        modified.

        The current principal is added to the owner (``write`` permission).

        Override to perform actions or post-process objects after their
        modification in storage.

        .. code-block:: python

            def update_object(self, obj, parent_id=None):
                obj = super().update_object(obj, parent_id)
                subject = f'Object {record[self.id_field]} was changed'
                send_email(subject)
                return object

        :param dict obj: object to store
        :param str parent_id: optional filter for parent id
        :returns: the updated object.
        :rtype: dict
        """
        parent_id = parent_id or self.parent_id
        object_id = obj[self.id_field]
        permissions = obj.pop(self.permissions_field, {})

        updated = self.storage.update(
            resource_name=self.resource_name,
            parent_id=parent_id,
            object_id=object_id,
            obj=obj,
            id_field=self.id_field,
            modified_field=self.modified_field,
        )

        perm_object_id = self.get_permission_object_id(object_id)
        self.permission.replace_object_permissions(perm_object_id, permissions)
        self._allow_write(perm_object_id)

        return self._annotate(updated, perm_object_id)

    def delete_object(self, obj, parent_id=None, last_modified=None):
        """Delete an object and its associated permissions.

        Override to perform actions or post-process objects after deletion
        from storage for example:

        .. code-block:: python

            def delete_object(self, obj):
                deleted = super().delete_object(obj)
                erase_media(obj)
                deleted['media'] = 0
                return deleted

        :param dict obj: the object to delete
        :param str parent_id: optional filter for parent id
        :returns: the deleted object.
        :rtype: dict
        """
        parent_id = parent_id or self.parent_id
        object_id = obj[self.id_field]
        perm_object_id = self.get_permission_object_id(object_id)

        self.permission.delete_object_permissions(perm_object_id)

        return self.storage.delete(
            resource_name=self.resource_name,
            parent_id=parent_id,
            object_id=object_id,
            id_field=self.id_field,
            modified_field=self.modified_field,
            deleted_field=self.deleted_field,
            last_modified=last_modified,
        )

    @property
    def collection_id(self):
        message = "`collection_id` is deprecated, use `resource_name` instead."
        warnings.warn(message, DeprecationWarning)
        return self.resource_name

    def get_records(self, *args, **kwargs):
        message = "`get_records()` is deprecated, use `get_objects()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.get_objects(*args, **kwargs)

    def delete_records(self, *args, **kwargs):
        message = "`delete_records()` is deprecated, use `delete_objects()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.delete_objects(*args, **kwargs)

    def get_record(self, record_id, parent_id=None):
        message = "`get_record()` is deprecated, use `get_object()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.get_object(object_id=record_id, parent_id=parent_id)

    def create_record(self, record, parent_id=None):
        message = "`create_record()` is deprecated, use `create_object()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.create_object(obj=record, parent_id=parent_id)

    def update_record(self, record, parent_id=None):
        message = "`update_record()` is deprecated, use `update_object()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.update_object(obj=record, parent_id=parent_id)

    def delete_record(self, record, parent_id=None, last_modified=None):
        message = "`delete_record()` is deprecated, use `delete_object()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.delete_object(obj=record, parent_id=parent_id, last_modified=last_modified)


class ShareableModel(Model):
    def __init__(self, *args, **kwargs):
        message = "`ShareableModel` is deprecated, use `Model` instead."
        warnings.warn(message, DeprecationWarning)
        super().__init__(*args, **kwargs)



from pyramid import httpexceptions

from kinto.core.errors import ERRORS

from . import BaseTest


class FilteringTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.validated = self.resource.request.validated
        self.patch_known_field.start()
        objects = [
            {"title": "MoFo", "status": 0, "favorite": True, "colors": ["blue", "red"]},
            {"title": "MoFo", "status": 1, "favorite": False, "colors": ["blue", "gray"]},
            {"title": "MoFo", "status": 2, "favorite": False, "sometimes": "available"},
            {"title": "MoFo", "status": 0, "favorite": False, "sometimes": None},
            {"title": "MoFo", "status": 1, "favorite": True, "fib": [1, 2, 3]},
            {"title": "MoFo", "status": 2, "favorite": False, "fib": [3, 5, 8]},
            {"title": "Foo", "status": 3, "favorite": False, "aliases": [{"ll": "ls -l"}]},
            {
                "title": "Bar",
                "status": 3,
                "favorite": False,
                "sometimes": "present",
                "aliases": [{"ll": "ls -l"}, {"rm": "rm -i"}],
            },
        ]
        for r in objects:
            self.model.create_object(r)

    def test_list_can_be_filtered_on_deleted_with_since(self):
        since = self.model.timestamp()
        r = self.model.create_object({})
        self.model.delete_object(r)
        self.validated["querystring"] = {"_since": since, "deleted": True}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 1)
        self.assertTrue(result["data"][0]["deleted"])

    def test_filter_on_id_is_supported(self):
        self.patch_known_field.stop()
        r = self.model.create_object({})
        r.pop(self.model.permissions_field)
        self.validated["querystring"] = {"id": "{}".format(r["id"])}
        result = self.resource.plural_get()
        self.assertEqual(result["data"][0], r)

    def test_list_cannot_be_filtered_on_deleted_without_since(self):
        r = self.model.create_object({})
        self.model.delete_object(r)
        self.validated["querystring"] = {"deleted": True}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 0)

    def test_filter_works_with_empty_list(self):
        self.resource.model.parent_id = "alice"
        self.validated["querystring"] = {"status": 1}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 0)

    def test_number_of_objects_matches_filter(self):
        self.validated["querystring"] = {"status": 1}
        self.resource.plural_head()
        headers = self.last_response.headers
        self.assertEqual(int(headers["Total-Objects"]), 2)

    def test_single_basic_filter_by_attribute(self):
        self.validated["querystring"] = {"status": 1}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 2)

    def test_filter_on_unknown_attribute_raises_error(self):
        self.patch_known_field.stop()
        self.validated["querystring"] = {"foo": 1}
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.plural_get)

    def test_filter_raises_error_if_last_modified_value_is_empty(self):
        self.validated["querystring"] = {"last_modified": ""}
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.plural_get)
        self.validated["querystring"] = {"_since": ""}
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.plural_get)
        self.validated["querystring"] = {"lt_last_modified": ""}
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.plural_get)

    def test_filter_errors_are_json_formatted(self):
        self.patch_known_field.stop()
        self.validated["querystring"] = {"foo": 1}
        try:
            self.resource.plural_get()
        except httpexceptions.HTTPBadRequest as e:
            error = e
        self.assertEqual(
            error.json,
            {
                "errno": ERRORS.INVALID_PARAMETERS.value,
                "message": "Unknown filter field 'foo'",
                "code": 400,
                "error": "Invalid parameters",
                "details": [
                    {
                        "description": "Unknown filter field 'foo'",
                        "location": "querystring",
                        "name": "foo",
                    }
                ],
            },
        )

    def test_regexp_is_strict_for_min_and_max(self):
        self.patch_known_field.stop()
        self.validated["querystring"] = {"madmax_status": 1}
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.plural_get)

    def test_double_basic_filter_by_attribute(self):
        self.validated["querystring"] = {"status": 1, "favorite": True}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 1)

    def test_string_filters_naively_by_value(self):
        self.validated["querystring"] = {"title": "MoF"}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 0)
        self.validated["querystring"] = {"title": "MoFo"}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 6)

    def test_not_string_filter(self):
        self.validated["querystring"] = {"like_title": 10}
        try:
            self.resource.plural_get()
        except httpexceptions.HTTPBadRequest as e:
            error = e
        self.assertEqual(
            error.json,
            {
                "code": 400,
                "details": [
                    {
                        "description": "Invalid value for like_title",
                        "location": "querystring",
                        "name": "like_title",
                    }
                ],
                "errno": 107,
                "error": "Invalid parameters",
                "message": "Invalid value for like_title",
            },
        )

    def test_string_filters_searching_by_value_not_matching(self):
        self.validated["querystring"] = {"like_title": "MoFoo"}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 0)

    def test_string_filters_searching_by_value_matching_many(self):
        self.validated["querystring"] = {"like_title": "Fo"}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 7)

    def test_string_filters_searching_by_value_matching_one(self):
        self.validated["querystring"] = {"like_title": "Bar"}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 1)

    def test_string_filters_searching_by_value_matching_vary_case(self):
        self.validated["querystring"] = {"like_title": "FoO"}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 1)

    def test_filter_considers_string_if_syntaxically_invalid(self):
        self.validated["querystring"] = {"status": "1.2.3"}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 0)

    def test_filter_does_not_fail_with_complex_type_syntax(self):
        self.validated["querystring"] = {"status": "(1,2,3)"}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 0)

    def test_different_value(self):
        self.validated["querystring"] = {"not_status": 2}
        result = self.resource.plural_get()
        values = [item["status"] for item in result["data"]]
        self.assertTrue(all([value != 2 for value in values]))

    def test_minimal_value(self):
        self.validated["querystring"] = {"min_status": 2}
        result = self.resource.plural_get()
        values = [item["status"] for item in result["data"]]
        self.assertTrue(all([value >= 2 for value in values]))

    def test_gt_value(self):
        self.validated["querystring"] = {"gt_status": 2}
        result = self.resource.plural_get()
        values = [item["status"] for item in result["data"]]
        self.assertTrue(all([value > 2 for value in values]))

    def test_maximal_value(self):
        self.validated["querystring"] = {"max_status": 2}
        result = self.resource.plural_get()
        values = [item["status"] for item in result["data"]]
        self.assertTrue(all([value <= 2 for value in values]))

    def test_lt_value(self):
        self.validated["querystring"] = {"lt_status": 2}
        result = self.resource.plural_get()
        values = [item["status"] for item in result["data"]]
        self.assertTrue(all([value < 2 for value in values]))

    def test_in_values(self):
        self.validated["querystring"] = {"in_status": [0, 1]}
        result = self.resource.plural_get()
        values = [item["status"] for item in result["data"]]
        self.assertEqual(sorted(values), [0, 0, 1, 1])

    def test_exclude_values(self):
        self.validated["querystring"] = {"exclude_status": [0]}
        result = self.resource.plural_get()
        values = [item["status"] for item in result["data"]]
        self.assertEqual(sorted(values), [1, 1, 2, 2, 3, 3])

    def test_has_values(self):
        self.validated["querystring"] = {"has_sometimes": True}
        result = self.resource.plural_get()
        values = [item["sometimes"] for item in result["data"]]
        assert None in values
        self.assertEqual(sorted([v for v in values if v]), ["available", "present"])

    def test_has_values_false(self):
        self.validated["querystring"] = {"has_sometimes": False}
        result = self.resource.plural_get()
        values = ["sometimes" in item for item in result["data"]]
        self.assertEqual(sorted(values), [False, False, False, False, False])

    def test_include_returns_400_if_value_has_wrong_type(self):
        self.validated["querystring"] = {"in_id": [0, 1]}
        with self.assertRaises(httpexceptions.HTTPBadRequest) as cm:
            self.resource.plural_get()
        self.assertIn("in_id", cm.exception.json["message"])

        self.validated["querystring"] = {"in_last_modified": ["a", "b"]}
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.plural_get)

    def test_exclude_returns_400_if_value_has_wrong_type(self):
        self.validated["querystring"] = {"exclude_id": [0, 1]}
        with self.assertRaises(httpexceptions.HTTPBadRequest) as cm:
            self.resource.plural_get()
        self.assertIn("exclude_id", cm.exception.json["message"])

        self.validated["querystring"] = {"exclude_last_modified": ["a", "b"]}
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.plural_get)

    def test_contains_can_filter_with_one_string(self):
        self.validated["querystring"] = {"contains_colors": ["red"]}
        result = self.resource.plural_get()
        values = [item["colors"] for item in result["data"]]
        assert len(values) == 1
        for value in values:
            assert "red" in value

    def test_contains_can_filter_with_one_object(self):
        self.validated["querystring"] = {"contains_aliases": [{"ll": "ls -l"}]}
        result = self.resource.plural_get()
        values = [item["aliases"] for item in result["data"]]
        assert len(values) == 2
        for value in values:
            assert {"ll": "ls -l"} in value

    def test_contains_can_filter_with_list_of_strings(self):
        self.validated["querystring"] = {"contains_colors": ["red", "blue"]}
        result = self.resource.plural_get()
        values = [item["colors"] for item in result["data"]]
        assert len(values) == 1
        for value in values:
            assert "red" in value and "blue" in value

    def test_contains_can_filter_with_an_integer(self):
        self.validated["querystring"] = {"contains_fib": [3]}
        result = self.resource.plural_get()
        values = [item["fib"] for item in result["data"]]
        assert len(values) == 2
        for value in values:
            assert 3 in value

    def test_contains_any_can_filter_with_a_list_of_strings(self):
        self.validated["querystring"] = {"contains_any_colors": ["red", "blue"]}
        result = self.resource.plural_get()
        values = [item["colors"] for item in result["data"]]
        assert len(values) == 2
        for value in values:
            assert "red" in value or "blue" in value

    def test_contains_any_can_filter_with_a_list_of_integers(self):
        self.validated["querystring"] = {"contains_any_fib": [3, 5]}
        result = self.resource.plural_get()
        values = [item["fib"] for item in result["data"]]
        assert len(values) == 2
        for value in values:
            assert 3 in value or 5 in value

    def test_contains_fails_on_a_non_sequence_object_value(self):
        self.validated["querystring"] = {"contains_favorite": [True]}
        result = self.resource.plural_get()
        values = result["data"]
        assert len(values) == 0

    def test_contains_any_fails_on_a_non_sequence_object_value(self):
        self.validated["querystring"] = {"contains_any_favorite": [True]}
        result = self.resource.plural_get()
        values = result["data"]
        assert len(values) == 0


class SubobjectFilteringTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.validated = self.resource.request.validated
        self.patch_known_field.start()
        for i in range(6):
            obj = {"party": {"candidate": "Marie", "voters": i}, "location": "Creuse"}
            self.model.create_object(obj)

    def test_objects_can_be_filtered_by_subobjects(self):
        self.validated["querystring"] = {"party.voters": 1}
        result = self.resource.plural_get()
        values = [item["party"]["voters"] for item in result["data"]]
        self.assertEqual(sorted(values), [1])

    def test_subobjects_filters_are_ignored_if_not_object(self):
        self.validated["querystring"] = {"location.city": "barcelona"}
        result = self.resource.plural_get()
        self.assertEqual(len(result["data"]), 0)

    def test_subobjects_filters_works_with_directives(self):
        self.validated["querystring"] = {"in_party.voters": [1, 2, 3]}
        result = self.resource.plural_get()
        values = [item["party"]["voters"] for item in result["data"]]
        self.assertEqual(sorted(values), [1, 2, 3])


class JSONFilteringTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.validated = self.resource.request.validated
        self.patch_known_field.start()
        objects = [
            {
                "id": "strawberry",
                "flavor": "strawberry",
                "orders": [],
                "attributes": {"ibu": 25, "seen_on": "2017-06-01"},
                "author": None,
            },
            {"id": "blueberry-1", "flavor": "blueberry", "orders": [1]},
            {"id": "blueberry-2", "flavor": "blueberry", "orders": ""},
            {"id": "raspberry-1", "flavor": "raspberry", "attributes": {}},
            {"id": "raspberry-2", "flavor": "raspberry", "attributes": []},
            {
                "id": "raspberry-3",
                "flavor": "raspberry",
                "attributes": {"ibu": 25, "seen_on": "2017-06-01", "price": 9.99},
            },
            {"id": "watermelon-1", "flavor": "watermelon", "author": "null"},
            {"id": "watermelon-2", "flavor": "watermelon", "author": 0},
        ]
        for r in objects:
            self.model.create_object(r)

    def test_filter_by_empty_array(self):
        self.validated["querystring"] = {"orders": []}
        result = self.resource.plural_get()
        assert len(result["data"]) == 1
        assert result["data"][0]["id"] == "strawberry"

    def test_filter_by_nonempty_array(self):
        self.validated["querystring"] = {"orders": [1]}
        result = self.resource.plural_get()
        assert len(result["data"]) == 1
        assert result["data"][0]["id"] == "blueberry-1"

    def test_filter_by_empty_object(self):
        self.validated["querystring"] = {"attributes": {}}
        result = self.resource.plural_get()
        assert len(result["data"]) == 1
        assert result["data"][0]["id"] == "raspberry-1"

    def test_filter_by_nonempty_object(self):
        self.validated["querystring"] = {"attributes": {"ibu": 25, "seen_on": "2017-06-01"}}
        result = self.resource.plural_get()
        assert len(result["data"]) == 1
        assert result["data"][0]["id"] == "strawberry"

    def test_filter_by_null(self):
        self.validated["querystring"] = {"author": None}
        result = self.resource.plural_get()
        assert len(result["data"]) == 1
        assert result["data"][0]["id"] == "strawberry"



import logging
import random

logger = logging.getLogger(__name__)


_HEARTBEAT_DELETE_RATE = 0.5
_HEARTBEAT_KEY = "__heartbeat__"
_HEARTBEAT_TTL_SECONDS = 3600


class CacheBase:
    def __init__(self, *args, **kwargs):
        self.prefix = kwargs["cache_prefix"]
        self.max_size_bytes = kwargs.get("cache_max_size_bytes")

    def initialize_schema(self, dry_run=False):
        """Create every necessary objects (like tables or indices) in the
        backend.

        This is executed when the ``kinto migrate`` command is run.

        :param bool dry_run: simulate instead of executing the operations.
        """
        raise NotImplementedError

    def flush(self):
        """Delete every values."""
        raise NotImplementedError

    def ttl(self, key):
        """Obtain the expiration value of the specified `key`.

        :param str key: key
        :returns: number of seconds or negative if no TTL.
        :rtype: float
        """
        raise NotImplementedError

    def expire(self, key, ttl):
        """Set the expiration value `ttl` for the specified `key`.

        :param str key: key
        :param float ttl: number of seconds
        """
        raise NotImplementedError

    def set(self, key, value, ttl):
        """Store a value with the specified `key`.

        :param str key: key
        :param str value: value to store
        :param float ttl: expire after number of seconds
        """
        raise NotImplementedError

    def get(self, key):
        """Obtain the value of the specified `key`.

        :param str key: key
        :returns: the stored value or None if missing.
        :rtype: str
        """
        raise NotImplementedError

    def delete(self, key):
        """Delete the value of the specified `key`.

        :param str key: key
        """
        raise NotImplementedError


def heartbeat(backend):
    def ping(request):
        """Test that cache backend is operational.

        :param request: current request object
        :type request: :class:`~pyramid:pyramid.request.Request`
        :returns: ``True`` is everything is ok, ``False`` otherwise.
        :rtype: bool
        """
        # No specific case for readonly mode because the cache should
        # continue to work in that mode.
        try:
            if random.SystemRandom().random() < _HEARTBEAT_DELETE_RATE:
                backend.delete(_HEARTBEAT_KEY)
                return backend.get(_HEARTBEAT_KEY) is None
            backend.set(_HEARTBEAT_KEY, "alive", _HEARTBEAT_TTL_SECONDS)
            return backend.get(_HEARTBEAT_KEY) == "alive"
        except Exception:
            logger.exception("Heartbeat Failure")
            return False

    return ping



from unittest import mock

from pyramid import testing

from kinto.core.permission import heartbeat
from kinto.core.storage import exceptions
from kinto.core.testing import DummyRequest


class PermissionTest:
    backend = None
    settings = {}

    def setUp(self):
        super().setUp()
        self.permission = self.backend.load_from_config(self._get_config())
        self.permission.initialize_schema()
        self.request = DummyRequest()
        self.client_error_patcher = []

    def _get_config(self):
        """Mock Pyramid config object.
        """
        config = testing.setUp()
        config.add_settings(self.settings)
        return config

    def tearDown(self):
        mock.patch.stopall()
        super().tearDown()
        self.permission.flush()

    def test_backend_error_is_raised_anywhere(self):
        for patch in self.client_error_patcher:
            patch.start()
        calls = [
            (self.permission.flush,),
            (self.permission.add_user_principal, "", ""),
            (self.permission.remove_user_principal, "", ""),
            (self.permission.get_user_principals, ""),
            (self.permission.add_principal_to_ace, "", "", ""),
            (self.permission.remove_principal_from_ace, "", "", ""),
            (self.permission.get_object_permission_principals, "", ""),
            (self.permission.get_object_permissions, ""),
            (self.permission.replace_object_permissions, "", {"write": []}),
            (self.permission.delete_object_permissions, ""),
            (self.permission.get_accessible_objects, []),
            (self.permission.get_authorized_principals, [("*", "read")]),
        ]
        for call in calls:
            self.assertRaises(exceptions.BackendError, *call)

    def test_initialize_schema_is_idempotent(self):
        self.permission.initialize_schema()
        self.permission.initialize_schema()  # not raising.

    def test_ping_returns_false_if_unavailable(self):
        ping = heartbeat(self.permission)
        for patch in self.client_error_patcher:
            patch.start()
        self.assertFalse(ping(self.request))

    def test_ping_returns_true_if_available(self):
        ping = heartbeat(self.permission)
        self.assertTrue(ping(self.request))

    def test_ping_returns_false_if_unavailable_in_readonly_mode(self):
        self.request.registry.settings["readonly"] = "true"
        ping = heartbeat(self.permission)
        with mock.patch.object(
            self.permission, "get_user_principals", side_effect=exceptions.BackendError("Boom!")
        ):
            self.assertFalse(ping(self.request))

    def test_ping_returns_true_if_available_in_readonly_mode(self):
        self.request.registry.settings["readonly"] = "true"
        ping = heartbeat(self.permission)
        self.assertTrue(ping(self.request))

    def test_ping_logs_error_if_unavailable(self):
        for patch in self.client_error_patcher:
            patch.start()
        ping = heartbeat(self.permission)

        with mock.patch("kinto.core.permission.logger.exception") as exc_handler:
            self.assertFalse(ping(self.request))

        self.assertTrue(exc_handler.called)

    def test_can_add_a_principal_to_a_user(self):
        user_id = "foo"
        principal = "bar"
        self.permission.add_user_principal(user_id, principal)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEqual(retrieved, {principal})

    def test_add_twice_a_principal_to_a_user_add_it_once(self):
        user_id = "foo"
        principal = "bar"
        self.permission.add_user_principal(user_id, principal)
        self.permission.add_user_principal(user_id, principal)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEqual(retrieved, {principal})

    def test_can_remove_a_principal_for_an_unknown_user(self):
        self.permission.remove_user_principal("foo", "bar")

    def test_can_remove_a_principal_for_a_user(self):
        user_id = "foo"
        principal = "bar"
        principal2 = "foobar"
        self.permission.add_user_principal(user_id, principal)
        self.permission.add_user_principal(user_id, principal2)
        self.permission.remove_user_principal(user_id, principal)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEqual(retrieved, {principal2})

    def test_can_remove_a_unexisting_principal_to_a_user(self):
        user_id = "foo"
        principal = "bar"
        principal2 = "foobar"
        self.permission.add_user_principal(user_id, principal2)
        self.permission.remove_user_principal(user_id, principal)
        self.permission.remove_user_principal(user_id, principal2)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEqual(retrieved, set())

    def test_can_remove_principal_from_every_users(self):
        user_id1 = "foo1"
        user_id2 = "foo2"
        principal1 = "bar"
        principal2 = "foobar"
        self.permission.add_user_principal(user_id1, principal1)
        self.permission.add_user_principal(user_id2, principal1)
        self.permission.add_user_principal(user_id2, principal2)
        self.permission.remove_principal(principal1)
        self.permission.remove_principal("unknown")

        retrieved = self.permission.get_user_principals(user_id1)
        self.assertEqual(retrieved, set())
        retrieved = self.permission.get_user_principals(user_id2)
        self.assertEqual(retrieved, {principal2})

    def test_cannot_remove_principal_from_objects(self):
        object_id = "/buckets/some-bucket"
        user_id = "user"
        principal = "readers-group"
        self.permission.add_principal_to_ace(object_id, "read", principal)
        self.permission.add_user_principal(user_id, principal)
        self.permission.remove_principal(principal)

        retrieved = self.permission.get_object_permissions(object_id)
        self.assertEqual(retrieved, {"read": set([principal])})

    def test_authenticated_is_returned_for_everybody(self):
        user_id = "foo"
        principal = "bar"
        self.permission.add_user_principal("system.Authenticated", principal)
        retrieved = self.permission.get_user_principals(user_id)
        self.assertEqual(retrieved, {principal})

    #
    # get_object_permission_principals()
    #

    def test_can_add_a_principal_to_an_object_permission(self):
        object_id = "foo"
        permission = "write"
        principal = "bar"
        self.permission.add_principal_to_ace(object_id, permission, principal)
        retrieved = self.permission.get_object_permission_principals(object_id, permission)
        self.assertEqual(retrieved, {principal})

    def test_add_twice_a_principal_to_an_object_permission_add_it_once(self):
        object_id = "foo"
        permission = "write"
        principal = "bar"
        self.permission.add_principal_to_ace(object_id, permission, principal)
        self.permission.add_principal_to_ace(object_id, permission, principal)
        retrieved = self.permission.get_object_permission_principals(object_id, permission)
        self.assertEqual(retrieved, {principal})

    def test_can_remove_a_principal_from_an_object_permission(self):
        object_id = "foo"
        permission = "write"
        principal = "bar"
        principal2 = "foobar"
        self.permission.add_principal_to_ace(object_id, permission, principal)
        self.permission.add_principal_to_ace(object_id, permission, principal2)
        self.permission.remove_principal_from_ace(object_id, permission, principal)
        retrieved = self.permission.get_object_permission_principals(object_id, permission)
        self.assertEqual(retrieved, {principal2})

    def test_principals_is_empty_if_no_permission(self):
        object_id = "foo"
        permission = "write"
        principal = "bar"
        self.permission.add_principal_to_ace(object_id, permission, principal)
        self.permission.remove_principal_from_ace(object_id, permission, principal)
        retrieved = self.permission.get_object_permission_principals(object_id, permission)
        self.assertEqual(retrieved, set())

    def test_can_remove_an_unexisting_principal_to_an_object_permission(self):
        object_id = "foo"
        permission = "write"
        principal = "bar"
        principal2 = "foobar"
        self.permission.add_principal_to_ace(object_id, permission, principal2)
        self.permission.remove_principal_from_ace(object_id, permission, principal)
        retrieved = self.permission.get_object_permission_principals(object_id, permission)
        self.assertEqual(retrieved, {principal2})

    #
    # check_permission()
    #

    def test_check_permission_returns_true_for_userid(self):
        object_id = "foo"
        permission = "write"
        principal = "bar"
        self.permission.add_principal_to_ace(object_id, permission, principal)
        check_permission = self.permission.check_permission({principal}, [(object_id, permission)])
        self.assertTrue(check_permission)

    def test_check_permission_returns_true_for_userid_group(self):
        object_id = "foo"
        permission = "write"
        group_id = "bar"
        user_id = "foobar"
        self.permission.add_user_principal(user_id, group_id)
        self.permission.add_principal_to_ace(object_id, permission, group_id)
        check_permission = self.permission.check_permission(
            {user_id, group_id}, [(object_id, permission)]
        )
        self.assertTrue(check_permission)

    def test_check_permission_returns_true_for_object_inherited(self):
        object_id = "foo"
        user_id = "bar"
        self.permission.add_principal_to_ace(object_id, "write", user_id)
        check_permission = self.permission.check_permission(
            {user_id}, [(object_id, "write"), (object_id, "read")]
        )
        self.assertTrue(check_permission)

    def test_check_permissions_handles_empty_set(self):
        principal = "bar"
        permits = self.permission.check_permission({principal}, [])
        self.assertFalse(permits)

    def test_check_permission_return_false_for_unknown_principal(self):
        object_id = "foo"
        permission = "write"
        principal = "bar"
        check_permission = self.permission.check_permission({principal}, [(object_id, permission)])
        self.assertFalse(check_permission)

    #
    # get_authorized_principals()
    #

    def test_get_authorized_principals_inherit_principals(self):
        object_id = "foo"
        user_id = "bar"
        self.permission.add_principal_to_ace(object_id, "write", user_id)
        principals = self.permission.get_authorized_principals(
            [(object_id, "write"), (object_id, "read")]
        )
        self.assertEqual(principals, {user_id})

    def test_get_authorized_principals_handles_empty_set(self):
        principals = self.permission.get_authorized_principals([])
        self.assertEqual(principals, set())

    #
    # get_accessible_objects()
    #

    def test_accessible_objects(self):
        self.permission.add_principal_to_ace("id1", "write", "user1")
        self.permission.add_principal_to_ace("id1", "object:create", "group")
        self.permission.add_principal_to_ace("id2", "read", "user1")
        self.permission.add_principal_to_ace("id2", "read", "user2")
        self.permission.add_principal_to_ace("id3", "write", "user2")
        per_object_ids = self.permission.get_accessible_objects(["user1", "group"])
        self.assertEqual(sorted(per_object_ids.keys()), ["id1", "id2"])
        self.assertEqual(per_object_ids["id1"], set(["write", "object:create"]))
        self.assertEqual(per_object_ids["id2"], set(["read"]))

    def test_accessible_objects_supports_empty_list(self):
        per_object_ids = self.permission.get_accessible_objects(["user1", "group"], [])
        self.assertEqual(per_object_ids, {})

    def test_accessible_objects_from_permission(self):
        self.permission.add_principal_to_ace("id1", "write", "user1")
        self.permission.add_principal_to_ace("id1", "read", "user1")
        self.permission.add_principal_to_ace("id1", "read", "group")
        self.permission.add_principal_to_ace("id2", "write", "user1")
        self.permission.add_principal_to_ace("id2", "read", "user2")
        self.permission.add_principal_to_ace("id2", "read", "group")
        self.permission.add_principal_to_ace("id3", "read", "user2")
        per_object_ids = self.permission.get_accessible_objects(
            ["user1", "group"], [("*", "read")]
        )
        self.assertEqual(sorted(per_object_ids.keys()), ["id1", "id2"])

    def test_accessible_objects_with_pattern(self):
        self.permission.add_principal_to_ace("/url1/id", "write", "user1")
        self.permission.add_principal_to_ace("/url2/id", "write", "user1")
        per_object_ids = self.permission.get_accessible_objects(["user1"], [("*url1*", "write")])
        self.assertEqual(sorted(per_object_ids.keys()), ["/url1/id"])

    def test_accessible_objects_with_pattern_matches_whole_id(self):
        self.permission.add_principal_to_ace("/url1/id", "write", "user1")
        self.permission.add_principal_to_ace("/url1/id/sub", "write", "user1")
        self.permission.add_principal_to_ace("/a/url1/id", "write", "user1")
        per_object_ids = self.permission.get_accessible_objects(
            ["user1"], [("/url1/*", "write")], with_children=False
        )
        self.assertEqual(sorted(per_object_ids.keys()), ["/url1/id"])

    def test_accessible_objects_several_bound_permissions(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/2", "read", "user1")
        self.permission.add_principal_to_ace("/url/_/id/2", "read", "user1")
        per_object_ids = self.permission.get_accessible_objects(
            ["user1"], [("/url/a/id/*", "read"), ("/url/a/id/*", "write")]
        )
        self.assertEqual(sorted(per_object_ids.keys()), ["/url/a/id/1", "/url/a/id/2"])

    def test_accessible_objects_without_match(self):
        self.permission.add_principal_to_ace("/url/a", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/b/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/2", "read", "user1")
        self.permission.add_principal_to_ace("/url/b/id/2", "read", "user1")
        per_object_ids = self.permission.get_accessible_objects(
            ["user1"],
            [
                ("/url/a", "write"),
                ("/url/a", "read"),
                ("/url/a/id/*", "write"),
                ("/url/a/id/*", "read"),
            ],
        )
        self.assertEqual(sorted(per_object_ids.keys()), ["/url/a", "/url/a/id/1", "/url/a/id/2"])

    #
    # get_object_permissions()
    #

    def test_object_permissions_return_all_object_acls(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user2")
        self.permission.add_principal_to_ace("/url/a/id/1", "read", "user3")
        self.permission.add_principal_to_ace("/url/a/id/1", "obj:del", "user1")
        self.permission.add_principal_to_ace("/url/a/id/1/sub", "create", "me")
        permissions = self.permission.get_object_permissions("/url/a/id/1")
        self.assertDictEqual(
            permissions, {"write": {"user1", "user2"}, "read": {"user3"}, "obj:del": {"user1"}}
        )

    def test_object_permissions_return_listed_object_acls(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user2")
        self.permission.add_principal_to_ace("/url/a/id/1", "read", "user3")
        self.permission.add_principal_to_ace("/url/a/id/1", "create", "user1")
        object_permissions = self.permission.get_object_permissions(
            "/url/a/id/1", ["write", "read"]
        )
        self.assertDictEqual(object_permissions, {"write": {"user1", "user2"}, "read": {"user3"}})

    def test_objects_permissions_returns_empty_if_unknown(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/3", "read", "user3")
        objects_permissions = self.permission.get_objects_permissions(
            ["/url/a/id/1", "/abc", "/url/a/id/3"]
        )
        self.assertEqual(objects_permissions, [{"write": {"user1"}}, {}, {"read": {"user3"}}])

    def test_object_permissions_return_empty_dict(self):
        self.assertDictEqual(self.permission.get_object_permissions("abc"), {})

    def test_replace_object_permission_replace_all_given_sets(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user2")
        self.permission.add_principal_to_ace("/url/a/id/1", "read", "user3")
        self.permission.add_principal_to_ace("/url/a/id/1", "update", "user1")
        self.permission.add_principal_to_ace("/url/a/id/1", "obj:del", "user1")

        self.permission.replace_object_permissions(
            "/url/a/id/1",
            {
                "write": ["user1"],
                "read": ["user2"],
                "update": [],
                "obj:del": ["user1"],
                "new": ["user3"],
            },
        )

        permissions = self.permission.get_object_permissions("/url/a/id/1")
        self.assertDictEqual(
            permissions,
            {"write": {"user1"}, "read": {"user2"}, "obj:del": {"user1"}, "new": {"user3"}},
        )

    def test_replace_object_permission_only_replace_given_sets(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user2")
        self.permission.add_principal_to_ace("/url/a/id/1", "read", "user3")
        self.permission.add_principal_to_ace("/url/a/id/1", "obj:del", "user1")

        self.permission.replace_object_permissions(
            "/url/a/id/1", {"write": ["user1"], "new": set(["user2"])}
        )

        permissions = self.permission.get_object_permissions("/url/a/id/1")
        self.assertDictEqual(
            permissions,
            {"write": {"user1"}, "read": {"user3"}, "new": {"user2"}, "obj:del": {"user1"}},
        )

    def test_replace_object_permission_supports_empty_existing_entries(self):
        self.permission.replace_object_permissions("/url/a/id/1", {"write": ["user1"]})
        permissions = self.permission.get_object_permissions("/url/a/id/1")
        self.assertDictEqual(permissions, {"write": {"user1"}})

    def test_replace_object_permission_supports_empty_input(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.replace_object_permissions("/url/a/id/1", {})
        permissions = self.permission.get_object_permissions("/url/a/id/1")
        self.assertDictEqual(permissions, {"write": {"user1"}})

    def test_replace_object_permission_supports_duplicated_entries(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.replace_object_permissions("/url/a/id/1", {"write": ["user1", "user1"]})
        permissions = self.permission.get_object_permissions("/url/a/id/1")
        self.assertDictEqual(permissions, {"write": {"user1"}})

    def test_replace_object_permission_supports_empty_list(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.replace_object_permissions("/url/a/id/1", {"write": set()})
        permissions = self.permission.get_object_permissions("/url/a/id/1")
        self.assertEqual(len(permissions), 0)

    def test_replace_object_permission_supports_empty_list_to_new_object(self):
        self.permission.replace_object_permissions("/url/a/id/1", {"write": set()})
        permissions = self.permission.get_object_permissions("/url/a/id/1")
        self.assertEqual(len(permissions), 0)

    def test_delete_object_permissions_remove_all_given_objects_acls(self):
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user2")
        self.permission.add_principal_to_ace("/url/a/id/1", "read", "user3")
        self.permission.add_principal_to_ace("/url/a/id/1", "create", "user1")
        self.permission.add_principal_to_ace("/url/a/id/2", "create", "user3")
        self.permission.add_principal_to_ace("/url/a/id/3", "create", "user4")

        self.permission.delete_object_permissions("/url/a/id/1", "/url/a/id/2")

        self.assertDictEqual(self.permission.get_object_permissions("/url/a/id/1"), {})
        self.assertDictEqual(self.permission.get_object_permissions("/url/a/id/2"), {})
        self.assertDictEqual(
            self.permission.get_object_permissions("/url/a/id/3"), {"create": {"user4"}}
        )

    def test_delete_object_permissions_supports_empty_list(self):
        self.permission.delete_object_permissions()  # Not failing

    def test_delete_object_permissions_supports_pattern_matching(self):
        self.permission.add_principal_to_ace("/url/b/id/1", "write", "user1")
        self.permission.add_principal_to_ace("/url/a/id/1", "write", "user2")
        self.permission.add_principal_to_ace("/url/a/id/1", "read", "user3")
        self.permission.add_principal_to_ace("/url/a/id/3", "create", "user4")

        self.permission.delete_object_permissions("/url/a*")

        self.assertDictEqual(self.permission.get_object_permissions("/url/a/id/1"), {})
        self.assertDictEqual(
            self.permission.get_object_permissions("/url/b/id/1"), {"write": {"user1"}}
        )
